#! /usr/bin/perl -w

# Directives.pm: handle directives in car files
# 
#    Copyright (C) 2005, Patrice Dumas
# 
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
use strict;

use CMZ::Conditionals;

package CMZ::Directives;

use vars qw(@ISA @EXPORT_OK $VERSION @EXPORT);

use Exporter;
$VERSION = '0.06';

@ISA = qw(Exporter);

@EXPORT = qw(%units);

our %units = ( 
	'title' => {
		'level'=>1,
	},
	'patch' => {
		'level' => 2,
		'name_length' => 16,
	},
	'deck' => {
		'level' => 3,
		'name_length' => 32,
	},
	'keep' => {
		'level' => 3,
		'name_length' => 32,
	}
);

my %conditionals = (
	'if' => 1,
	'else' => 1,
	'elseif' => 1,
	'endif' => 0,
);

my %misc = (
	'seq' => {'special'=>
		['dateqq','timeqq','qftitlch','qftitle','versqq','authqq','vidqq']},
# undocumented
	'cde' => {'special'=>
		['dateqq','timeqq','qftitlch','qftitle','versqq','authqq','vidqq']},
	'self' => 1,
);

sub is_seq($)
{
	my $cmd = shift;
	return 1 if ($cmd eq 'seq' or $cmd eq 'cde');
}

sub conditionals($)
{
	my $cond = shift;
	return ($conditionals{$cond}) if (exists($conditionals{$cond}));
	return undef;
}

my %directives = ();
foreach my $directive ((keys(%units), keys(%conditionals), keys(%misc)))
{
	$directives{$directive} = 1;
}

sub parse_directive($$)
{
	my $line_orig = shift;
	my $verbose = shift;
	return 0 if (!defined($line_orig));
	my $line = $line_orig;
	$line =~ s/\..*//;
	$line =~ s/(,\s*)*$//;
	my $directive = undef;
	if ($line =~ s/^\+_*([a-z]+)\s*//i)
	{
		$directive = $1;
		print STDERR "Maybe directive: $directive\n" if ($verbose>1);
		return 0 if (!$directives{lc($directive)});	
	}
	return 0 if (!defined($directive));
	# remove all spaces in line before analysing. Seems like it is 
	# what cmz does
	$line =~ s/\s*//g;
	print STDERR "Directive $directive -> $line\n" if ($verbose > 1);
	
	return ({'directive'=>$directive,'level'=>$units{lc($directive)}->{'level'}})
		if (lc($directive) eq 'title');
	my $retval = {'directive'=>$directive};
	if ($units{lc($directive)})
	{
		my $name_length = '';
		$name_length = $units{lc($directive)}->{'name_length'} 
			if (defined($units{lc($directive)}->{'name_length'}));
		my $name;
		if ($line =~ s/^,\s*(\*?[\w\-\.\$]{1,$name_length})//)
		{
			$name = $1;
		}
		if (!defined($name))
		{
			print STDERR "Directive lacking name: \n$line_orig";
			return 0;
		}
		$retval->{'lang'} = '';
		$retval->{'level'} = $units{lc($directive)}->{'level'};
		$retval->{'name'} = $name;
		return $retval if ($line =~ /^\s*$/);
		if ($line =~ s/^(\w+)//)
		{
			print STDERR "Truncated too long name: \n$line_orig";
		}
		return $retval if ($line =~ /^\s*$/);
		$line =~ s/^\s*//;
		my ($if,$if_val,$t);
		($if,$if_val,$t,$line) = parse_if_t($line,$line_orig);
		$retval->{'if'} = $if if (defined($if));
		$retval->{'if_val'} = $if_val if (defined($if_val));
		if (defined($t))
		{
			foreach my $type(@$t)
			{
				if (lc($type) eq 'pass')
				{
					$retval->{'pass'} = 1;
				}
				else
				{
					$retval->{'lang'} = $type;
				}
			}
		}
		#print STDERR "".Data::Dumper->Dump([$retval])."\n";
	}
	elsif(exists($conditionals{lc($directive)}))
	{
		if ($conditionals{lc($directive)})
		{
		#	print STDERR "Conditional: $directive\n" if ($verbose > 1);
		#	if ($line =~ s/^\s*,\s*([^\s]+.*)\s*//)
		#	{
		#		my $if = $1;
		#		$retval = {'directive'=>$directive, 'if'=>$if,
		#			'if_val'=>CMZ::Conditionals::parse($if)};
		#	}
			$line =~ s/^\s*(,\s*)+\s*/,/;
			unless ($line =~ s/^\s*,\s*$//)
			{
				if ($line !~ /^,\s*if\s*=/i)
				{
					$line =~ s/^,/,if=/;
				}
				my ($if,$if_val,$t);
				($if,$if_val,$t,$line) = parse_if_t($line,$line_orig);
				$retval->{'if'} = $if if (defined($if));
				$retval->{'if_val'} = $if_val if (defined($if_val));
				if (defined($t))
				{
					foreach my $type(@$t)
					{
						print STDERR "Ignoring T=$type in +$directive: $line_orig";
					}
				}
			}
		}
		else
		{
			$retval = {'directive'=>$directive};
		}
	}
	elsif(lc($directive) eq 'self')
		#and ($line =~ /^\s*,\s*(if|t)\s*=\s*/i))
	{
		$line =~ s/^\s*(,\s*)+\s*/,/;
		if ($line =~ /,\s*(if|t)\s*=/i)
		{
			$line =~ s/^.*?(,\s*(if|t)\s*=)/$1/i;
			my ($if,$if_val,$t);
			($if,$if_val,$t,$line) = parse_if_t($line,$line_orig);
			$retval->{'if'} = $if if (defined($if));
			$retval->{'if_val'} = $if_val if (defined($if_val));
			if (defined($t))
			{
				foreach my $type(@$t)
				{
					if (lc($type) eq 'pass')
					{
						$retval->{'pass'} = 1;
					}
					else
					{
						print STDERR "bad T=$type in +SELF:\n$line_orig";
					}
				}
			}
		}
		else
		{
			$line = '';
		}
	}
	elsif(is_seq(lc($directive)))
	{
		my @seq = ();
		my $special_seq;
		if ($line =~ /^,\s*(\w+)/)
		{
			$special_seq = lc($1);
			my @specials = grep {/^$special_seq/} @{$misc{lc($directive)}->{'special'}};
			if (@specials)
			{
				$special_seq = $specials[0];
				print STDERR "special($directive): $special_seq\n" if ($verbose > 1);
				$line =~ s/^,\s*(\w+)\s*//;
				my $special_arg;
				while ($line =~ /^,\s*([Nn]|[tT])\s*=\s*/)
				{
					if ($line =~ /^,\s*[Nn]\s*=\s*(\d+)\s*/)
					{
						$line =~ s/^,\s*[Nn]\s*=\s*(\d+)\s*//;
						$special_arg = $1;
					}
					elsif ($line =~ s/^,\s*[tT]\s*=\s*(\w+)\s*//)
					{
						if (lc($1) eq 'pass')
						{
							$retval->{'pass'} = 1;
						}
					}
				}
				$retval->{'special'} = $special_seq;
				$retval->{'special_arg'} = $special_arg;
			}
			else
			{
				$special_seq = undef;
			}
		}
		unless ($special_seq)
		{
			while ($line !~ /^,\s*if\s*=\s*/i and
				$line !~ /^,\s*t\s*=\s*(\w+)\s*/i)
			{
				if ($line =~ s/^,\s*([\w\-\.\$]*)\s*//i)
				{
					my $seq = $1;
					next if ($1 eq '');
					push @seq, $seq;
				}
				else
				{
					if ($line !~ /^\s*$/)
					{
						print STDERR "Strange line:\n$line_orig";
					}
					last;
				}
			}
			if (!@seq)
			{
				print STDERR "seq without name:\n$line_orig";
			}
			else
			{
				$retval->{'seq'} = [@seq];
			}
			return $retval if ($line =~ /^\s*$/);
			$line =~ s/^\s*//;
			my ($if,$if_val,$t);
			($if,$if_val,$t,$line) = parse_if_t($line,$line_orig);
			$retval->{'if'} = $if if (defined($if));
			$retval->{'if_val'} = $if_val if (defined($if_val));
			foreach my $type(@$t)
			{
				if (lc($type) eq 'noinclude')
				{
					$retval->{'noinclude'} = $type; 
				}
				else
				{
					print STDERR "Strange T=$type for +SEQ\n";
				}
			}
		}
	}
	if ($line !~ /^\s*$/)
	{
		chomp $line;
		print STDERR "Garbage at end of line ($line) for:\n $line_orig";
	}
	return $retval;
}

sub parse_if_t($$)
{
	my $line = shift;
	my $line_orig = shift;
	my $if;
	my $t;
	my @if_args = ();
	my @t_args = ();
	while ($line =~ /^,\s*if\s*=\s*/i or 
		$line =~ /^,\s*t\s*=\s*/i)
	{
		#if ($line =~ s/^,\s*t\s*=\s*(\w[\w,]*)\s*//i)
		#{
	#		my @t = split /\s*,\s*/, $1;
	#		$t = \@t;
#			next;
#		}
		my ($if_arg,$t_arg);
		if ($line =~ s/^,\s*((if\s*=\s*(.*?))|(t\s*=\s*(\w[\w,]*?)))\s*(,\s*(if|t)\s*=\s*)/$6/i)
		{
			my $if_or_t = $1;
			my $if = $2;
			$if_arg = $3;
			my $t = $4;
			$t_arg = $5;
			#print STDERR "--> $if_or_t\n";
			#print STDERR "-->$line_orig\n";
			#print STDERR "--> $if_or_t ! $if !  $if_arg ! $t ! $t_arg ! $6\n";
		}
		elsif ($line =~ s/^,\s*((if\s*=\s*(.+))|(t\s*=\s*(\w[\w,]*?)))\s*$//i)
		{
			my $if_or_t = $1;
			my $if = $2;
			$if_arg = $3;
			my $t = $4;
			$t_arg = $5;
			#print STDERR "end -->$if_or_t\n";
			#print STDERR "e-->$line_orig\n";
			#print STDERR "e--> $if_or_t ! $if !  $if_arg ! $t ! $t_arg\n";
		}
		$if_arg = undef if (defined($if_arg) and $if_arg =~ /^\s*$/);
		$t_arg = undef if (defined($t_arg) and $t_arg =~ /^\s*$/);

		if (defined($if_arg))
		{
			push @if_args,$if_arg;
			#last;
		}
		elsif (defined($t_arg))
		{
			push @t_args,split (/\s*,\s*/,$t_arg);
			#print STDERR "debug: @t_args | $t_arg\n";
			#last;
		}
		else
		{
			print STDERR "Bad T=,IF= section:\n$line_orig";
		}
		#print STDERR "Remaining: $line\n";
	}
	if (@if_args)
	{
		$if = shift(@if_args);
		if (@if_args)
		{
			$if = "($if)";
			foreach my $if_arg(@if_args)
			{
				$if .= "&($if_arg)";
			}
		}
	}
	if (defined($if))
	{
		return ($if,CMZ::Conditionals::parse($if),\@t_args,$line);
	}
	else
	{
		return (undef,undef,\@t_args,$line);
	}
}

1;
